package jcifs.auth;

import java.io.IOException;

import java.net.UnknownHostException;

import java.security.Principal;

import java.util.Enumeration;

import javax.servlet.Filter;
import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;

import jcifs.Config;
import jcifs.UniAddress;

import jcifs.auth.ntlm.NtlmToken;
import jcifs.auth.ntlm.Type1Token;
import jcifs.auth.ntlm.Type2Token;
import jcifs.auth.ntlm.Type3Token;

import jcifs.http.NtlmHttpServletRequest;

import jcifs.smb.NtlmPasswordAuthentication;
import jcifs.smb.SmbAuthException;
import jcifs.smb.SmbSession;

import jcifs.util.Base64;

public class HttpFilter implements Filter {

    public static final String PRINCIPAL = "jcifs.auth.principal";

    private static final int UNKNOWN = -1;

    private static final int NEGOTIATE = 0;

    private static final int NTLM = 1;

    private static final int BASIC = 2;

    private FilterConfig config;

    private UniAddress domainController;

    private boolean enableBasic;

    private boolean insecureBasic;

    private String realm;

    public void init(FilterConfig config) throws ServletException {
        this.config = config;
        Config.setProperty("jcifs.smb.client.soTimeout", "300000");
        Config.setProperty("jcifs.netbios.cachePolicy", "600");
        Enumeration parameters = config.getInitParameterNames();
        while (parameters.hasMoreElements()) {
            String parameter = (String) parameters.nextElement();
            if (parameter.startsWith("jcifs.")) {
                Config.setProperty(parameter,
                        config.getInitParameter(parameter));
            }
        }
        enableBasic = Boolean.valueOf(
                Config.getProperty("jcifs.http.enableBasic")).booleanValue();
        insecureBasic = Boolean.valueOf(
                Config.getProperty("jcifs.http.insecureBasic")).booleanValue();
        realm = Config.getProperty("jcifs.http.basicRealm");
        if (realm == null) realm = "jCIFS";
    }

    private synchronized UniAddress getDomainController() {
        if (domainController != null) return domainController;
        String domainController =
                Config.getProperty("jcifs.http.domainController");
        if (domainController == null) {
            domainController = Config.getProperty("jcifs.smb.client.domain");
        }
        if (domainController == null) return null;
        try {
            return (this.domainController =
                    UniAddress.getByName(domainController));
        } catch (UnknownHostException ex) {
            return null;
        }
    }

    public void destroy() {
        config = null;
    }

    public FilterConfig getFilterConfig() {
        return config;
    }

    public void setFilterConfig(FilterConfig config) {
        try {
            init(config);
        } catch (ServletException ex) {
            throw new IllegalStateException("Unable to initialize.");
        }
    }

    public void doFilter(ServletRequest req, ServletResponse resp,
            FilterChain chain) throws IOException, ServletException {
        HttpServletRequest request = (HttpServletRequest) req;
        HttpServletResponse response = (HttpServletResponse) resp;
        String authorization = request.getHeader("Authorization");
        Principal principal = null;
        switch (getAuthenticationType(authorization)) {
        case NEGOTIATE:
            principal = doNegotiate(Base64.decode(authorization.substring(10)),
                    request, response, chain);
            break;
        case NTLM:
            principal = doNtlm(Base64.decode(authorization.substring(5)),
                    request, response, chain);
            break;
        case BASIC:
            principal = doBasic(Base64.decode(authorization.substring(6)),
                    request, response, chain);
            break;
        default:
            principal = doUnknown(request, response, chain);
        }
        if (response.isCommitted()) return;
        if (principal != null) {
            request.getSession().setAttribute(PRINCIPAL, principal);
            chain.doFilter(new NtlmHttpServletRequest(request, principal),
                    response);
            return;
        }
        //response.setHeader("WWW-Authenticate", "Negotiate");
        response.addHeader("WWW-Authenticate", "NTLM");
        if (enableBasic && (insecureBasic || request.isSecure())) {
            response.addHeader("WWW-Authenticate", "Basic realm=\"" +
                    realm + "\"");
        }
        response.setHeader("Connection", "close");
        response.setStatus(HttpServletResponse.SC_UNAUTHORIZED);
        response.flushBuffer();
    }

    private Principal doNegotiate(byte[] message, HttpServletRequest request,
            HttpServletResponse response, FilterChain chain)
                    throws ServletException, IOException {
        // not yet
        return null;
    }

    private Principal doNtlm(byte[] message, HttpServletRequest request,
            HttpServletResponse response, FilterChain chain)
                    throws ServletException, IOException {
        TokenFactory tokenFactory = TokenFactory.getInstance();
        NtlmToken token = (NtlmToken) tokenFactory.createToken(message);
        if (token instanceof Type1Token) {
            Type1Token type1 = (Type1Token) token;
            Type2Token type2 = new Type2Token(type1);
            type2.setChallenge(SmbSession.getChallenge(getDomainController()));
            response.setHeader("WWW-Authenticate", "NTLM " +
                    Base64.encodeBytes(tokenFactory.createMaterial(type2)));
            response.setContentLength(0);
            response.setStatus(HttpServletResponse.SC_UNAUTHORIZED);
            response.flushBuffer();
            return null;
        } else if (token instanceof Type3Token) {
            Type3Token type3 = (Type3Token) token;
            NtlmPasswordAuthentication ntlm = new NtlmPasswordAuthentication(
                    type3.getDomain(), type3.getUser(), type3.getLMResponse(),
                            type3.getNTResponse());
            try {
                SmbSession.logon(getDomainController(), ntlm);
            } catch (SmbAuthException sae) {
                return null;
            }
            request.getSession().setAttribute("NtlmHttpAuth", ntlm);
            return ntlm;
        } else {
            return null;
        }
    }

    private Principal doBasic(byte[] message, HttpServletRequest request,
            HttpServletResponse response, FilterChain chain)
                    throws ServletException, IOException {
        String authorization = new String(message, "UTF-8");
        int index = authorization.indexOf(':');
        String user = (index != -1) ? authorization.substring(0, index) :
                authorization;
        String password = (index != -1) ? authorization.substring(index + 1) :
                "";
        index = user.indexOf('\\');
        if (index == -1) index = user.indexOf('/');
        String domain = (index != -1) ? user.substring(0, index) :
                Config.getProperty("jcifs.smb.client.domain");
        user = (index != -1) ? user.substring(index + 1) : user;
        NtlmPasswordAuthentication ntlm =
                new NtlmPasswordAuthentication(domain, user, password);
        try {
            SmbSession.logon(getDomainController(), ntlm);
        } catch (SmbAuthException sae) {
            return null;
        }
        request.getSession().setAttribute("NtlmHttpAuth", ntlm);
        return ntlm;
    }

    private Principal doUnknown(HttpServletRequest request,
            HttpServletResponse response, FilterChain chain)
                    throws ServletException, IOException {
        HttpSession session = request.getSession(false);
        return (session != null) ? (Principal) session.getAttribute(PRINCIPAL) :
                null;
    }

    private int getAuthenticationType(String authorization) {
        if (authorization == null) return UNKNOWN;
        int index = authorization.indexOf(' ');
        if (index == -1) return UNKNOWN;
        authorization = authorization.substring(0, index);
        if ("Negotiate".equals(authorization)) return NEGOTIATE;
        if ("NTLM".equals(authorization)) return NTLM;
        if ("Basic".equals(authorization)) return BASIC;
        return UNKNOWN;
    }

}

