package jcifs.auth;

import java.net.UnknownHostException;

import java.util.List;
import java.util.Map;
import java.util.Vector;

import javax.security.auth.Subject;

import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.callback.NameCallback;
import javax.security.auth.callback.PasswordCallback;
import javax.security.auth.callback.TextOutputCallback;

import javax.security.auth.login.FailedLoginException;
import javax.security.auth.login.LoginException;

import jcifs.UniAddress;

import jcifs.smb.NtlmPasswordAuthentication;
import jcifs.smb.SmbException;
import jcifs.smb.SmbSession;

/**
 * Implementation of a JAAS-compatible pluggable authentication module.
 * This differs from the Sun-provided NTLoginModule in that
 * <ol>
 * <li>It authenticates arbitrary users (rather than just the user logged
 * in at the console); and</li>
 * <li>It is platform-independent and will work on non-Windows boxes.</li>
 * </ol>
 */
public class LoginModule implements javax.security.auth.spi.LoginModule {

    private static final String USERNAME = "javax.security.auth.login.name";

    private static final String PASSWORD = "javax.security.auth.login.password";

    private Subject subject;

    private CallbackHandler handler;

    private Map sharedState;

    private NtlmPasswordAuthentication authentication;

    private Map options;

    private UniAddress domainController;

    private String defaultDomain;

    private String bannerMessage;

    private boolean tryFirstPass;

    private boolean useFirstPass;

    private boolean storePass;

    private boolean clearPass;

    public void initialize(Subject subject, CallbackHandler handler,
            Map sharedState, Map options) {
        this.subject = subject;
        this.handler = handler;
        this.sharedState = sharedState;
        this.options = options;
        try {
            domainController = UniAddress.getByName((String)
                    options.get("domainController"));
        } catch (UnknownHostException ex) {
            throw new IllegalStateException("Unable to locate host: " +
                    options.get("domainController"));
        }
        defaultDomain = (String) options.get("defaultDomain");
        bannerMessage = (String) options.get("bannerMessage");
        tryFirstPass = Boolean.valueOf((String)
                options.get("tryFirstPass")).booleanValue();
        useFirstPass = Boolean.valueOf((String)
                options.get("useFirstPass")).booleanValue();
        storePass = Boolean.valueOf((String)
                options.get("storePass")).booleanValue();
        clearPass = Boolean.valueOf((String)
                options.get("clearPass")).booleanValue();
    }

    public boolean login() throws LoginException {
        try {
            String username = null;
            char[] password = null;
            if (useFirstPass || tryFirstPass) {
                username = (String) sharedState.get(USERNAME);
                password = (char[]) sharedState.get(PASSWORD);
                authentication = logon(username, password);
                if (authentication != null) return true;
                if (useFirstPass) {
                    throw new FailedLoginException("Unable to authenticate " +
                            username + ".");
                }
            }
            if (handler == null) {
                throw new LoginException("No callback handler provided to " +
                        "gather credentials.");
            }
            List callbackList = new Vector();
            if (bannerMessage != null) {
                callbackList.add(new TextOutputCallback(
                        TextOutputCallback.INFORMATION, bannerMessage));
            }
            callbackList.add(new NameCallback("Username: "));
            callbackList.add(new PasswordCallback("Password: ", false));
            Callback[] callbacks = (Callback[])
                    callbackList.toArray(new Callback[0]);
            try {
                handler.handle(callbacks);
            } catch (Exception ex) {
                throw new LoginException("Unable to log in: " + ex);
            }
            for (int i = callbacks.length - 1; i >= 0; i--) {
                if (callbacks[i] instanceof PasswordCallback) {
                    char[] pwd = ((PasswordCallback)
                            callbacks[i]).getPassword();
                    if (pwd == null) pwd = new char[0];
                    password = new char[pwd.length];
                    System.arraycopy(pwd, 0, password, 0, pwd.length);
                    ((PasswordCallback) callbacks[i]).clearPassword();
                    for (int j = pwd.length - 1; j >= 0; j--) pwd[j] = ' ';
                } else if (callbacks[i] instanceof NameCallback) {
                    username = ((NameCallback) callbacks[i]).getName();
                }
            }
            authentication = logon(username, password);
            if (authentication != null) return true;
            throw new FailedLoginException("Unable to authenticate " +
                    username + ".");
        } finally {
            if (clearPass) {
                sharedState.remove(USERNAME);
                char[] pwd = (char[]) sharedState.get(PASSWORD);
                if (pwd != null) {
                    sharedState.remove(PASSWORD);
                    for (int i = pwd.length - 1; i >= 0; i--) pwd[i] = ' ';
                }
            }
        }
    }

    public boolean commit() throws LoginException {
        if (authentication == null) return false;
        subject.getPrincipals().add(authentication);
        return true;
    }

    public boolean abort() throws LoginException {
        if (authentication == null) return false;
        subject.getPrincipals().remove(authentication);
        logoff();
        authentication = null;
        return true;
    }

    public boolean logout() throws LoginException {
        if (authentication == null) return true;
        if (!subject.isReadOnly()) {
            subject.getPrincipals().remove(authentication);
        }
        logoff();
        authentication = null;
        return true;
    }

    protected NtlmPasswordAuthentication logon(String username,
            char[] password) {
        int index = (username != null) ? username.indexOf('\\') : -1;
        if (index == -1) {
            index = (username != null) ? username.indexOf('/') : -1;
        }
        String domain = (index != -1) ? username.substring(0, index) : null;
        if (index != -1) username = username.substring(index + 1);
        if (domain == null) domain = defaultDomain;
        if (password == null) password = new char[0];
        NtlmPasswordAuthentication authentication =
                new NtlmPasswordAuthentication(domain, username,
                        new String(password));
        try {
            SmbSession.logon(domainController, authentication);
        } catch (SmbException ex) {
            authentication = null;
        }
        for (int i = password.length - 1; i >= 0; i--) password[i] = ' ';
        return authentication;
    }

    protected void logoff() throws LoginException {
       /*  Currently does nothing.  I had this doing
        *
        * SmbTransport.getSmbTransport(domainController, 0).getSmbSession(
        *       authentication).logoff(false);
        *
        * But due to the package-private methods this required the class to
        * be in the jcifs.smb package.  I don't know if the above is really
        * necessary anyways; I'm guessing not.
        */
    }

}
