package jcifs.auth.ntlm;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;

import java.util.Arrays;

import jcifs.Config;

import jcifs.auth.Token;
import jcifs.auth.TokenFactory;
import jcifs.auth.UnrecognizedTokenException;

public class NtlmTokenFactory extends TokenFactory {

    public static final String NTLM_ALGORITHM = "NTLMSSP";

    private static final String OEM_ENCODING =
            Config.getProperty("jcifs.smb.client.codepage",
                    Config.getProperty("jcifs.encoding",
                            System.getProperty("file.encoding")));

    public String getAlgorithm() {
        return NTLM_ALGORITHM;
    }

    public Token createToken(byte[] material)
            throws UnrecognizedTokenException {
        try {
            if (!"NTLMSSP\u0000".equals(
                    new String(material, 0, 8, "US-ASCII"))) {
                throw new UnrecognizedTokenException("Not NTLMSSP material.");
            }
        } catch (UnsupportedEncodingException ex) {
            throw new IllegalStateException();
        }
        try {
            switch (readULong(material, 8)) {
            case 1:
                return createType1Token(material);
            case 2:
                return createType2Token(material);
            case 3:
                return createType3Token(material);
            default:
                throw new UnrecognizedTokenException("Unrecognized type.");
            }
        } catch (UnrecognizedTokenException ex) {
            throw ex;
        } catch (IOException ex) {
            throw new UnrecognizedTokenException(ex.getMessage());
        }
    }

    public byte[] createMaterial(Token token)
            throws UnrecognizedTokenException {
        try {
            if (token instanceof Type1Token) {
                return createType1Material((Type1Token) token);
            } else if (token instanceof Type2Token) {
                return createType2Material((Type2Token) token);
            } else if (token instanceof Type3Token) {
                return createType3Material((Type3Token) token);
            } else {
                throw new UnrecognizedTokenException("Not an NTLMSSP token.");
            }
        } catch (UnrecognizedTokenException ex) {
            throw ex;
        } catch (IOException ex) {
            throw new UnrecognizedTokenException(ex.getMessage());
        }
    }

    private static byte[] createType1Material(Type1Token token)
            throws IOException {
        ByteArrayOutputStream type1 = new ByteArrayOutputStream();
        type1.write("NTLMSSP\u0000".getBytes("US-ASCII"));
        writeULong(type1, 1);
        int flags = token.getFlags();
        String suppliedDomain = token.getSuppliedDomain();
        byte[] domain = new byte[0];
        if (suppliedDomain != null && suppliedDomain.length() != 0) {
            flags |= NtlmToken.NTLMSSP_NEGOTIATE_OEM_DOMAIN_SUPPLIED;
            domain = suppliedDomain.toUpperCase().getBytes(OEM_ENCODING);
        } else {
            flags &= (NtlmToken.NTLMSSP_NEGOTIATE_OEM_DOMAIN_SUPPLIED ^
                    0xffffffff);
        }
        String suppliedWorkstation = token.getSuppliedWorkstation();
        byte[] workstation = new byte[0];
        if (suppliedWorkstation != null && suppliedWorkstation.length() != 0) {
            flags |= NtlmToken.NTLMSSP_NEGOTIATE_OEM_WORKSTATION_SUPPLIED;
            workstation =
                    suppliedWorkstation.toUpperCase().getBytes(OEM_ENCODING);
        } else {
            flags &= (NtlmToken.NTLMSSP_NEGOTIATE_OEM_WORKSTATION_SUPPLIED ^
                    0xffffffff);
        }
        writeULong(type1, flags);
        int length = domain.length;
        writeUShort(type1, length);
        writeUShort(type1, length);
        // IE clients put the workstation buffer before the domain
        // (i.e., reversed), so for compatibility we will too.
        length = workstation.length;
        writeULong(type1, 32 + length);
        writeUShort(type1, length);
        writeUShort(type1, length);
        writeULong(type1, 32);
        type1.write(workstation);
        type1.write(domain);
        return type1.toByteArray();
    }

    private static byte[] createType2Material(Type2Token token)
            throws IOException {
        ByteArrayOutputStream type2 = new ByteArrayOutputStream();
        type2.write("NTLMSSP\u0000".getBytes("US-ASCII"));
        writeULong(type2, 2);
        int flags = token.getFlags();
        String targetString = token.getTarget();
        byte[] target = new byte[0];
        if ((flags & (NtlmToken.NTLMSSP_TARGET_TYPE_DOMAIN |
                NtlmToken.NTLMSSP_TARGET_TYPE_SERVER |
                        NtlmToken.NTLMSSP_TARGET_TYPE_SHARE)) != 0) {
            if (targetString != null && targetString.length() != 0) {
                target = (flags & NtlmToken.NTLMSSP_NEGOTIATE_UNICODE) != 0 ?
                        targetString.getBytes("UnicodeLittleUnmarked") :
                        targetString.toUpperCase().getBytes(OEM_ENCODING);
            } else {
                flags &= (0xffffffff ^ (NtlmToken.NTLMSSP_TARGET_TYPE_DOMAIN |
                        NtlmToken.NTLMSSP_TARGET_TYPE_SERVER |
                                NtlmToken.NTLMSSP_TARGET_TYPE_SHARE));
            }
        }
        int length = target.length;
        writeUShort(type2, length);
        writeUShort(type2, length);
        writeULong(type2, 32);
        writeULong(type2, flags);
        byte[] challenge = new byte[8];
        byte[] bytes = token.getChallenge();
        if (bytes != null) {
            System.arraycopy(bytes, 0, challenge, 0, Math.min(bytes.length, 8));
        }
        type2.write(challenge);
        type2.write(target);
        return type2.toByteArray();
    }

    private static byte[] createType3Material(Type3Token token)
            throws IOException {
        ByteArrayOutputStream type3 = new ByteArrayOutputStream();
        type3.write("NTLMSSP\u0000".getBytes("US-ASCII"));
        writeULong(type3, 3);
        byte[] lmResponse = token.getLMResponse();
        if (lmResponse == null) lmResponse = new byte[0];
        int lmResponseLength = lmResponse.length;
        byte[] ntResponse = token.getNTResponse();
        if (ntResponse == null) ntResponse = new byte[0];
        int ntResponseLength = ntResponse.length;
        int flags = token.getFlags();
        String domainString = token.getDomain();
        byte[] domain = new byte[0];
        if (domainString != null && domainString.length() != 0) {
            domain = (flags & NtlmToken.NTLMSSP_NEGOTIATE_UNICODE) != 0 ?
                    domainString.getBytes("UnicodeLittleUnmarked") :
                            domainString.toUpperCase().getBytes(OEM_ENCODING);
        }
        String userString = token.getUser();
        byte[] user = new byte[0];
        if (userString != null && userString.length() != 0) {
            user = (flags & NtlmToken.NTLMSSP_NEGOTIATE_UNICODE) != 0 ?
                    userString.getBytes("UnicodeLittleUnmarked") :
                            userString.toUpperCase().getBytes(OEM_ENCODING);
        }
        String workstationString = token.getWorkstation();
        byte[] workstation = new byte[0];
        if (workstationString != null && workstationString.length() != 0) {
            workstation = (flags & NtlmToken.NTLMSSP_NEGOTIATE_UNICODE) != 0 ?
                    workstationString.getBytes("UnicodeLittleUnmarked") :
                            workstationString.toUpperCase().getBytes(
                                    OEM_ENCODING);
        }
        int domainOffset = 64;
        int domainLength = domain.length;
        int userOffset = domainOffset + domainLength;
        int userLength = user.length;
        int workstationOffset = userOffset + userLength;
        int workstationLength = workstation.length;
        int lmResponseOffset = workstationOffset + workstationLength;
        int ntResponseOffset = lmResponseOffset + lmResponseLength;
        writeUShort(type3, lmResponseLength);
        writeUShort(type3, lmResponseLength);
        writeULong(type3, lmResponseOffset);
        writeUShort(type3, ntResponseLength);
        writeUShort(type3, ntResponseLength);
        writeULong(type3, ntResponseOffset);
        writeUShort(type3, domainLength);
        writeUShort(type3, domainLength);
        writeULong(type3, domainOffset);
        writeUShort(type3, userLength);
        writeUShort(type3, userLength);
        writeULong(type3, userOffset);
        writeUShort(type3, workstationLength);
        writeUShort(type3, workstationLength);
        writeULong(type3, workstationOffset);
        // session key (empty, placed at the end for compatibility)
        writeUShort(type3, 0);
        writeUShort(type3, 0);
        writeULong(type3, ntResponseOffset + ntResponseLength);
        writeULong(type3, flags);
        type3.write(domain);
        type3.write(user);
        type3.write(workstation);
        type3.write(lmResponse);
        type3.write(ntResponse);
        return type3.toByteArray();
    }

    private static byte[] readSecurityBuffer(byte[] src, int index)
            throws IOException {
        int length = readUShort(src, index);
        int offset = readULong(src, index + 4);
        byte[] buffer = new byte[length];
        System.arraycopy(src, offset, buffer, 0, length);
        return buffer;
    }

    private static int readUShort(byte[] src, int index) throws IOException {
        return (src[index] & 0xff) | ((src[index + 1] & 0xff) << 8);
    }

    private static int readULong(byte[] src, int index) throws IOException {
        return (src[index] & 0xff) |
                ((src[index + 1] & 0xff) << 8) |
                ((src[index + 2] & 0xff) << 16) |
                ((src[index + 3] & 0xff) << 24);
    }

    private static void writeULong(OutputStream output, int ulong)
            throws IOException {
        output.write(ulong & 0xff);
        output.write((ulong >> 8) & 0xff);
        output.write((ulong >> 16) & 0xff);
        output.write((ulong >> 24) & 0xff);
    }

    private static void writeUShort(OutputStream output, int ushort)
            throws IOException {
        output.write(ushort & 0xff);
        output.write((ushort >> 8) & 0xff);
    }

    private static Type1Token createType1Token(byte[] material)
            throws IOException {
        int flags = readULong(material, 12);
        String suppliedDomain = null;
        if ((flags & NtlmToken.NTLMSSP_NEGOTIATE_OEM_DOMAIN_SUPPLIED) != 0) {
            byte[] domain = readSecurityBuffer(material, 16);
            suppliedDomain = new String(domain, OEM_ENCODING);
        }
        String suppliedWorkstation = null;
        if ((flags &
                NtlmToken.NTLMSSP_NEGOTIATE_OEM_WORKSTATION_SUPPLIED) != 0) {
            byte[] workstation = readSecurityBuffer(material, 24);
            suppliedWorkstation = new String(workstation, OEM_ENCODING);
        }
        return new Type1Token(flags, suppliedDomain, suppliedWorkstation);
    }

    private static Type2Token createType2Token(byte[] material)
            throws IOException {
        int flags = readULong(material, 20);
        String target = null;
        byte[] bytes = readSecurityBuffer(material, 12);
        if (bytes.length != 0) {
            target = new String(bytes,
                    ((flags & NtlmToken.NTLMSSP_NEGOTIATE_UNICODE) != 0) ?
                            "UnicodeLittleUnmarked" : OEM_ENCODING);
        }
        byte[] challenge = new byte[8];
        System.arraycopy(material, 24, challenge, 0, 8);
        if (Arrays.equals(challenge, new byte[8])) challenge = null;
        return new Type2Token(flags, challenge, target);
    }

    private static Type3Token createType3Token(byte[] material)
            throws IOException {
        byte[] lmResponse = readSecurityBuffer(material, 12);
        int lmResponseOffset = readULong(material, 16);
        byte[] ntResponse = readSecurityBuffer(material, 20);
        int ntResponseOffset = readULong(material, 24);
        byte[] domain = readSecurityBuffer(material, 28);
        int domainOffset = readULong(material, 32);
        byte[] user = readSecurityBuffer(material, 36);
        int userOffset = readULong(material, 40);
        byte[] workstation = readSecurityBuffer(material, 44);
        int workstationOffset = readULong(material, 48);
        if (lmResponseOffset == 52 || ntResponseOffset == 52 ||
                domainOffset == 52 || userOffset == 52 ||
                        workstationOffset == 52) {
            // non-conformant win9x
            int flags = NtlmToken.NTLMSSP_NEGOTIATE_NTLM |
                    NtlmToken.NTLMSSP_REQUEST_TARGET |
                            NtlmToken.NTLMSSP_NEGOTIATE_OEM;
            return new Type3Token(flags, lmResponse, ntResponse,
                    new String(domain, OEM_ENCODING),
                            new String(user, OEM_ENCODING),
                                    new String(workstation, OEM_ENCODING));
        }
        int flags = readULong(material, 60);
        String charset = ((flags & NtlmToken.NTLMSSP_NEGOTIATE_UNICODE) != 0) ?
                "UnicodeLittleUnmarked" : OEM_ENCODING;
        return new Type3Token(flags, lmResponse, ntResponse,
                new String(domain, charset), new String(user, charset),
                        new String(workstation, charset));
    }

}
