package jcifs.auth.spnego;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;

import java.math.BigInteger;

import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.StringTokenizer;
import java.util.Vector;

import jcifs.auth.Token;
import jcifs.auth.TokenFactory;
import jcifs.auth.UnrecognizedTokenException;

public class SpnegoTokenFactory extends TokenFactory {

    public static final String SPNEGO_ALGORITHM = "SPNEGO";

    public static final String KERBEROS_MECHANISM = "1.2.840.113554.1.2.2";

    public static final String MS_KERBEROS_MECHANISM = "1.2.840.48018.1.2.2";

    public static final String NTLMSSP_MECHANISM = "1.3.6.1.4.1.311.2.2.10";

    public static final String SPNEGO_MECHANISM = "1.3.6.1.5.5.2";

    private static final byte[] SPNEGO_OID = createOID(SPNEGO_MECHANISM);

    private static final int CONSTRUCTED = 0x20;

    private static final int APPLICATION = 0x40;
    private static final int CONTEXT_SPECIFIC = 0x80;

    private static final int BIT_STRING = 0x03;
    private static final int OCTET_STRING = 0x04;
    private static final int OID = 0x06;
    private static final int ENUMERATED = 0x0a;
    private static final int SEQUENCE = 0x10;

    private static final int NEG_TOKEN_INIT = CONSTRUCTED | CONTEXT_SPECIFIC |
            0x00;
    private static final int NEG_TOKEN_TARG = CONSTRUCTED | CONTEXT_SPECIFIC |
            0x01;

    private static final int MECHANISM_TYPES = CONSTRUCTED | CONTEXT_SPECIFIC |
            0x00;
    private static final int REQUEST_FLAGS = CONSTRUCTED | CONTEXT_SPECIFIC |
            0x01;
    private static final int MECHANISM_MATERIAL = CONSTRUCTED |
            CONTEXT_SPECIFIC | 0x02;
    private static final int MECHANISM_LIST_MIC = CONSTRUCTED |
            CONTEXT_SPECIFIC | 0x03;

    private static final int RESULT = CONSTRUCTED | CONTEXT_SPECIFIC |
            0x00;
    private static final int MECHANISM = CONSTRUCTED | CONTEXT_SPECIFIC |
            0x01;

    public String getAlgorithm() {
        return SPNEGO_ALGORITHM;
    }

    public Token createToken(byte[] material)
            throws UnrecognizedTokenException {
        ByteArrayInputStream input = new ByteArrayInputStream(material);
        if (input.read() == (CONSTRUCTED | APPLICATION)) {
            int length = readLength(input);
            if (input.read() != OID) {
                throw new UnrecognizedTokenException("No object identifier.");
            }
            length = readLength(input);
            if (!Arrays.equals(SPNEGO_OID, readOID(input))) {
                throw new UnrecognizedTokenException("Not SPNEGO material.");
            }
        }
        switch (input.read()) {
        case NEG_TOKEN_INIT:
            return readInitiatorToken(input);
        case NEG_TOKEN_TARG:
            return readTargetToken(input);
        default:
            throw new UnrecognizedTokenException("Unknown material type.");
        }
    }

    public byte[] createMaterial(Token token)
            throws UnrecognizedTokenException {
        if (token instanceof InitiatorToken) {
            return createInitiatorMaterial((InitiatorToken) token);
        } else if (token instanceof TargetToken) {
            return createTargetMaterial((TargetToken) token);
        } else {
            throw new UnrecognizedTokenException("Not a SPNEGO token.");
        }
    }

    public static String getMechanism(byte[] material)
            throws UnrecognizedTokenException {
        try {
            return createMechanism(findOID(material));
        } catch (Exception ex) {
            throw new UnrecognizedTokenException(ex.getMessage());
        }
    }

    private static byte[] createInitiatorMaterial(InitiatorToken token) {
        ByteStack stack = new ByteStack();
        int length = 0;
        byte[] bytes = token.getMechanismListMIC();
        if (bytes != null) {
            stack.put(bytes);
            length = bytes.length;
            bytes = createHeader(OCTET_STRING, length);
            stack.put(bytes);
            stack.put(createHeader(MECHANISM_LIST_MIC, length + bytes.length));
        }
        bytes = token.getMaterial();
        if (bytes != null) {
            stack.put(bytes);
            length = bytes.length;
            bytes = createHeader(OCTET_STRING, length);
            stack.put(bytes);
            stack.put(createHeader(MECHANISM_MATERIAL, length + bytes.length));
        }
        bytes = createFlags(token.getFlags());
        if (bytes != null) {
            stack.put(bytes);
            length = bytes.length;
            bytes = createHeader(BIT_STRING, length);
            stack.put(bytes);
            stack.put(createHeader(REQUEST_FLAGS, length + bytes.length));
        }
        length = 0;
        bytes = token.getMaterial();
        if (bytes != null) {
            bytes = findOID(bytes);
            String[] mechanisms = token.getMechanisms();
            if (mechanisms != null) {
                for (int i = mechanisms.length - 1; i >= 0; i--) {
                    byte[] oid = createOID(mechanisms[i]);
                    if (bytes == null || !Arrays.equals(bytes, oid)) {
                        stack.put(oid);
                        length += oid.length;
                        oid = createHeader(OID, oid.length);
                        stack.put(oid);
                        length += oid.length;
                    }
                }
            }
            if (bytes != null) {
                stack.put(bytes);
                length += bytes.length;
                bytes = createHeader(OID, bytes.length);
                stack.put(bytes);
                length += bytes.length;
            }
            bytes = createHeader(CONSTRUCTED | SEQUENCE, length);
            stack.put(bytes);
            stack.put(createHeader(MECHANISM_TYPES, bytes.length + length));
        }
        stack.put(createHeader(CONSTRUCTED | SEQUENCE, stack.size()));
        stack.put(createHeader(NEG_TOKEN_INIT, stack.size()));
        stack.put(SPNEGO_OID);
        stack.put(createHeader(OID, SPNEGO_OID.length));
        stack.put(createHeader(CONSTRUCTED | APPLICATION, stack.size()));
        return stack.get();
    }

    private static byte[] createTargetMaterial(TargetToken token) {
        ByteStack stack = new ByteStack();
        int length = 0;
        byte[] bytes = token.getMechanismListMIC();
        if (bytes != null) {
            stack.put(bytes);
            length = bytes.length;
            bytes = createHeader(OCTET_STRING, length);
            stack.put(bytes);
            stack.put(createHeader(MECHANISM_LIST_MIC, length + bytes.length));
        }
        bytes = token.getMaterial();
        if (bytes != null) {
            stack.put(bytes);
            length = bytes.length;
            bytes = createHeader(OCTET_STRING, length);
            stack.put(bytes);
            stack.put(createHeader(MECHANISM_MATERIAL, length + bytes.length));
        }
        bytes = createOID(token.getMechanism());
        if (bytes != null) {
            stack.put(bytes);
            length = bytes.length;
            bytes = createHeader(OID, length);
            stack.put(bytes);
            stack.put(createHeader(MECHANISM, length + bytes.length));
        }
        int result = token.getResult();
        switch (result) {
        case TargetToken.ACCEPT_COMPLETED:
        case TargetToken.ACCEPT_INCOMPLETE:
            if (token.getMechanism() != null) {
                stack.put(new byte[] { (byte) result });
                bytes = createHeader(ENUMERATED, 1);
                stack.put(bytes);
                stack.put(createHeader(RESULT, bytes.length + 1));
            }
            break;
        case TargetToken.REJECTED:
            if (token.getMechanism() != null) {
                throw new IllegalStateException(
                        "Supplied mechanism with rejection.");
            }
            stack.put(new byte[] { (byte) result });
            bytes = createHeader(ENUMERATED, 1);
            stack.put(bytes);
            stack.put(createHeader(RESULT, bytes.length + 1));
            break;
        default:
            throw new IllegalArgumentException("Invalid result code.");
        }
        stack.put(createHeader(CONSTRUCTED | SEQUENCE, stack.size()));
        stack.put(createHeader(NEG_TOKEN_TARG, stack.size()));
        return stack.get();
    }

    private static InitiatorToken readInitiatorToken(
            ByteArrayInputStream input) {
        int length = readLength(input);
        if (input.read() != (CONSTRUCTED | SEQUENCE)) {
            throw new IllegalArgumentException("No content.");
        }
        byte[] mechanismMaterial = null;
        String[] mechanisms = null;
        int flags = 0;
        byte[] mic = null;
        length = readLength(input);
        while (length > 0) {
            int type = input.read();
            int partLength = readLength(input);
            switch (input.read()) {
            case MECHANISM_TYPES:
                mechanisms = readMechanisms(input);
                break;
            case REQUEST_FLAGS:
                flags = readRequestFlags(input);
                break;
            case MECHANISM_MATERIAL:
                mechanismMaterial = readOctetString(input);
                break;
            case MECHANISM_LIST_MIC:
                mic = readOctetString(input);
                break;
            default:
                throw new IllegalStateException("Unrecognized content.");
            }
            length -= partLength;
        }
        return new InitiatorToken(mechanismMaterial, mechanisms, flags, mic);
    }

    private static TargetToken readTargetToken(ByteArrayInputStream input) {
        int length = readLength(input);
        if (input.read() != (CONSTRUCTED | SEQUENCE)) {
            throw new IllegalArgumentException("No content.");
        }
        byte[] mechanismMaterial = null;
        String mechanism = null;
        int result = TargetToken.ACCEPT_COMPLETED;
        byte[] mic = null;
        length = readLength(input);
        while (length > 0) {
            int type = input.read();
            int partLength = readLength(input);
            switch (input.read()) {
            case RESULT:
                result = readResult(input);
                break;
            case MECHANISM:
                mechanism = createMechanism(readOID(input));
                break;
            case MECHANISM_MATERIAL:
                mechanismMaterial = readOctetString(input);
                break;
            case MECHANISM_LIST_MIC:
                mic = readOctetString(input);
                break;
            default:
                throw new IllegalStateException("Unrecognized content.");
            }
            length -= partLength;
        }
        return new TargetToken(mechanismMaterial, mechanism, result, mic);
    }

    private static String[] readMechanisms(ByteArrayInputStream input) {
        if (input.read() != (CONSTRUCTED | SEQUENCE)) {
            throw new IllegalStateException("No mechanism list content.");
        }
        int length = readLength(input);
        byte[] bytes = new byte[length];
        length = input.read(bytes, 0, length);
        if (length != bytes.length) {
            throw new IllegalStateException("Unable to fully read mechanisms.");
        }
        input = new ByteArrayInputStream(bytes);
        List mechanisms = new Vector();
        while (input.available() != 0) {
            mechanisms.add(createMechanism(readOID(input)));
        }
        return (String[]) mechanisms.toArray(new String[0]);
    }

    private static int readRequestFlags(ByteArrayInputStream input) {
        if (input.read() != BIT_STRING) {
            throw new IllegalStateException("No request flag content.");
        }
        int length = readLength(input);
        byte[] bytes = new byte[length];
        length = input.read(bytes, 0, length);
        if (length != 2) {
            throw new IllegalStateException("Unable to fully read flags.");
        }
        // we know the length is 2, and the first octet is 1 (number of shifts).
        return bytes[1] >>> 1;
    }

    private static int readResult(ByteArrayInputStream input) {
        if (input.read() != ENUMERATED) {
            throw new IllegalStateException("No result content.");
        }
        int length = readLength(input);
        // length will always be 1
        return input.read();
    }

    private static byte[] readOctetString(ByteArrayInputStream input) {
        if (input.read() != OCTET_STRING) {
            throw new IllegalStateException("Not an octet string.");
        }
        int length = readLength(input);
        byte[] bytes = new byte[length];
        length = input.read(bytes, 0, length);
        if (length != bytes.length) {
            throw new IllegalStateException("Unable to fully read content.");
        }
        return bytes;
    }

    private static byte[] readOID(ByteArrayInputStream input) {
        if (input.read() != OID) {
            throw new IllegalArgumentException("Not an OID.");
        }
        int length = readLength(input);
        byte[] bytes = new byte[length];
        input.read(bytes, 0, length);
        return bytes;
    }

    private static byte[] findOID(byte[] material) {
        if (material == null) return null;
        ByteArrayInputStream input = new ByteArrayInputStream(material);
        if (input.read() != (CONSTRUCTED | APPLICATION)) return null;
        readLength(input);
        if (input.read() != OID) return null;
        int length = readLength(input);
        byte[] bytes = new byte[length];
        length = input.read(bytes, 0, length);
        if (length != bytes.length) {
            throw new IllegalStateException("Unable to fully read OID.");
        }
        return bytes;
    }

    private static String createMechanism(byte[] oid) {
        if (oid == null) return null;
        int valueX = oid[0];
        int value1;
        int value2;
        if (valueX > 79) {
            value1 = 2;
            value2 = valueX - 80;
        } else if (valueX > 39) {
            value1 = 1;
            value2 = valueX - 40;
        } else {
            value1 = 0;
            value2 = valueX;
        }
        StringBuffer buffer = new StringBuffer();
        buffer.append(value1).append('.').append(value2);
        int length = oid.length;
        for (int i = 1; i < length; i++) {
            buffer.append('.');
            BigInteger valueN = new BigInteger("0");
            while ((oid[i] & 0x80) != 0) {
                valueN = valueN.shiftLeft(7).or(
                        new BigInteger(String.valueOf(oid[i] & 0x7f)));
                i++;
            }
            valueN = valueN.shiftLeft(7).or(
                    new BigInteger(String.valueOf(oid[i] & 0x7f)));
            buffer.append(valueN.toString());
        }
        return buffer.toString();
    }

    private static byte[] createOID(String mechanism) {
        if (mechanism == null) return null;
        ByteArrayOutputStream collector = new ByteArrayOutputStream();
        StringTokenizer tokenizer = new StringTokenizer(mechanism, ".");
        int value1 = Integer.parseInt(tokenizer.nextToken());
        int value2 = Integer.parseInt(tokenizer.nextToken());
        // byte 1 is (40 * first component) + second component
        collector.write((40 * value1) + value2);
        // rest of bytes are base-128 representation of each component
        // high bit set for all but the last byte of each block
        while (tokenizer.hasMoreTokens()) {
            BigInteger valueN = new BigInteger(tokenizer.nextToken());
            int length = valueN.bitLength();
            if (length % 7 != 0) {
                length = length + 7 - (length % 7);
            }
            for (int i = length; i > 0;) {
                int value = (i != 7) ? 0x80 : 0x00;
                do {
                    if (valueN.testBit(--i)) value |= 0x01 << (i % 7);
                } while (i % 7 != 0);
                collector.write(value);
            }
        }
        return collector.toByteArray();
    }

    private static byte[] createFlags(int flags) {
        if (flags == 0) return null;
        // DER-encoded bit string; we know a zero will need to be right-padded.
        flags <<= 1;
        byte[] bytes = new byte[2];
        bytes[0] = (byte) 0x01;
        bytes[1] = (byte) (flags & 0x0ff);
        return bytes;
    }

    private static byte[] createHeader(int type, int length) {
        byte[] header;   
        if (length < 0x80) {
            header = new byte[2];
            header[1] = (byte) (length & 0x7f);
        } else {
            byte[] lengthBytes =
                    new BigInteger(String.valueOf(length)).toByteArray();
            if (lengthBytes[0] == (byte) 0x00) {
                header = new byte[1 + lengthBytes.length];
                System.arraycopy(lengthBytes, 1, header, 2,
                        lengthBytes.length - 1);
            } else {
                header = new byte[2 + lengthBytes.length];
                System.arraycopy(lengthBytes, 0, header, 2, lengthBytes.length);
            }
            header[1] = (byte) ((lengthBytes.length & 0xff) | 0x80);
        }
        header[0] = (byte) (type & 0xff);
        return header;
    }

    private static int readLength(ByteArrayInputStream input) {
        int length = input.read();
        if ((length & 0x80) != 0) {
            byte[] lengthBytes = new byte[(length & 0x7f) + 1];
            // add a zero at the front to ensure sign is correct
            input.read(lengthBytes, 1, lengthBytes.length - 1);
            return new BigInteger(lengthBytes).intValue();
        } else {
            return length;
        }
    }

    private static class ByteStack {

        private int size;

        private final List content = new Vector();

        public int size() {
            return size;
        }

        public void put(byte[] bytes) {
            content.add(0, bytes);
            size += bytes.length;
        }

        public byte[] get() {
            byte[] bytes = new byte[size];
            int index = 0;
            Iterator blocks = content.iterator();
            while (blocks.hasNext()) {
                byte[] block = (byte[]) blocks.next();
                System.arraycopy(block, 0, bytes, index, block.length);
                index += block.length;
            }
            return bytes;
        }

    }

}
